/***************************************************************************************
 *																													*
 *	SCOPE_PLUGIN.V:	Top level plug-in for the GR8BUS master board for GR8BIT digial	*
 *																			oscilloscope implementation	*
 *																						http://www.gr8bit.ru	*
 *	Copyleft (cl) 2013 by Eugeny Brychkov																*
 *																													*
 *									License and T&C															*
 *																													*
 * You can use and modify this software's source and executable, given two					*
 * conditions are met: (a) you keep this Copyleft statement in the source code,			*
 * and you do not distribute software and resulting executables on the commercial		*
 * basis and for the purposes, not related to GR8BIT platform and/or MSX standard,		*
 * without written permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)			*
 *																													*
 ***************************************************************************************
 *			File name: SCOPE_PLUGIN.V																		*
 **************************************************************************************/

 /* Release date: July 01, 2013
  
	Techical information:
	- this configuration is to be directly plugged into the main configuration of the
	  GR8BUS master board (uses its signals directly)
	- use "include" directive to attach to the project, and modify/comment BD/DIO
	  wires accordingly
	- sampling time quantum depends on the PLL0_sampleCLK cycle time. For the DRAM chip usually
	  used for V9938 VDP minimal delay between activation or /RAS and /CAS is 12ns, thus we used
	  sampling clock equal to 6*VDPCLK (cycle of 7.759ns). However it could be safe to use 2*VDPCLK
	  clock (23.28ns) because VDP's datasheet states 40ns minimum for /RAS to /CAS delay. You can
	  change the PLL clock factor in the MegaWizard Plug-In Manager in the file "scope_mem_sp.v".
*/

//******************************************************************
// Definitions
//******************************************************************
// Globals
wire	PLL0_sampleCLK;				// main sample clock from PLL
wire negclock = ~PLL0_sampleCLK;	// negated PLL clock, for preparation of data for RAM write

// Host interface
reg [15:0] host_read_address;		// address to read data from RAM buffer
reg bclock;								// is used to identify rising edge of the system clock (I_BCLOCK)

// Sampling process
reg sampling_start;				// flag from host to start sampling
reg [15:0] write_record;		// contains address of the next RAM address to write sampled data to
reg [31:0] cur_time;				// current time relative to the start of the sampling process, in sample clock cycles
reg [31:0] prev_time;			// the time when previous sampling occured, used to calculate duration of state
reg [5:0] time_delta;			// register holding duration of the state to be written to RAM
reg [3:0] prev_control_lines;	// previous state of the control lines to derive lines' change flag(s)
reg [3:0] cur_control_lines;	// current contol lines latched on the positive edge of the sample clock
reg [7:0] data_lines;			// latched scope's data lines to write to the RAM buffer

wire [3:0] control_lines_changed_state = cur_control_lines[3:0] ^ prev_control_lines[3:0];		// flags if respective control lines state has changed
wire control_lines_changed =	control_lines_changed_state[3] | control_lines_changed_state[2] |
										control_lines_changed_state[1] | control_lines_changed_state[0]; 	// flag if any control list state has changed

wire [31:0] time_diff = cur_time[31:0] - prev_time[31:0];	// time difference between current and previous sampling stages
wire sampling_end = (write_record[15:0] == 16'h_83_ff);		// end of sampling period = buffer is full

wire [17:0] record = { time_delta[5:0], cur_control_lines[3:0], data_lines[7:0] };		// record to write to RAM buffer
wire [3:0] control_lines_live = { CLK7, CLK6, CLK5, CLK4 };										// control lines, physical connection
wire [7:0] data_lines_live = { ID0, SDA, AL[5:0] };												// data lines, physical connection

reg [7:0] data_lines_CLK4_ne;	// four registers holding latched data on respective _asynchronous_ control signal activation
reg [7:0] data_lines_CLK5_ne;
reg [7:0] data_lines_CLK6_ne;
reg [7:0] data_lines_CLK7_ne;

wire	ram_wr;						// RAM write signal, 1 for sample data chunk write, otherwise 0
wire [17:0] ram_out;				// RAM output data bus to present to the host
wire [15:0] ram_address = sampling_start ? write_record[15:0] : host_read_address[15:0];	// RAM chip address, equals to sampling record #
																														// if sampling started, otherwise to host
																														// requested address

//******************************************************************
// Generation of the clock using VDPCLK signal
//******************************************************************
scope_pll pll_module (
	.inclk0	(I_BVDPCLK),
	.c0		(PLL0_sampleCLK)		// used factor 6, equals to 128.880 MHz (7.759ns)
);

//******************************************************************
// Address lines state latch: value is only meaningful on
// negative edges of the control signals (RAS and CAS)
//******************************************************************
always @(negedge CLK4) data_lines_CLK4_ne[7:0] <= data_lines_live[7:0];		// CAS0
always @(negedge CLK5) data_lines_CLK5_ne[7:0] <= data_lines_live[7:0];		// CAS1
always @(negedge CLK6) data_lines_CLK6_ne[7:0] <= data_lines_live[7:0];		// RAS
always @(negedge CLK7) data_lines_CLK7_ne[7:0] <= data_lines_live[7:0];		// BLNK

//******************************************************************
// Control lines state latch (for stability)
//******************************************************************
always @(posedge PLL0_sampleCLK or negedge BRESET_N or negedge sampling_start) begin
	if(!BRESET_N || !sampling_start) cur_control_lines[3:0] <= 4'b1111;
	else cur_control_lines[3:0] <= control_lines_live[3:0];
end

//******************************************************************
// Sampling process
//******************************************************************
always @(posedge negclock or negedge BRESET_N or negedge sampling_start) begin
	if(!BRESET_N || !sampling_start) begin
		// if reset condition or sampling was not started/reset, reset all variables
		cur_time[31:0] <= 32'b0;
		prev_time[31:0] <= 32'b0;
		write_record[15:0] <= 16'b0;
		prev_control_lines[3:0] <= 4'b1111;
		ram_wr <= 1'b0;
	end
	else begin
		if(!sampling_end) begin		// only proceed with next sampling stage if RAM buffer was not reached
			if(control_lines_changed || time_diff[5:0]=={6{1'b1}}) begin		// if there was change in control line for previous period, or counter max
				if(control_lines_changed_state[0] && !cur_control_lines[0]) data_lines[7:0] <= data_lines_CLK4_ne[7:0];		// /CAS0, higherst priority
					else	if(control_lines_changed_state[1] && !cur_control_lines[1]) data_lines[7:0] <= data_lines_CLK5_ne[7:0];
							else	if(control_lines_changed_state[2] && !cur_control_lines[2]) data_lines[7:0] <= data_lines_CLK6_ne[7:0];
									else	if(control_lines_changed_state[3] && !cur_control_lines[3]) data_lines[7:0] <= data_lines_CLK7_ne[7:0];	// BLNK, lowest priority
				write_record[15:0] <= write_record[15:0] + 16'b1;	// next RAM address
				time_delta[5:0] <= time_diff[5:0];						// set delta with previous time (6 bits only!)
				prev_time[31:0] <= cur_time[31:0];						// set prev time to current time
				prev_control_lines[3:0] <= cur_control_lines[3:0];
				ram_wr <= 1'b1;												// force write to RAM buffer
			end
			else	ram_wr <= 1'b0;											// otherwise no write to RAM
			cur_time[31:0] <= cur_time[31:0] + 32'b1;					// next time tick
		end
		else ram_wr <= 1'b0;
	end
end

//******************************************************************
// RAM buffer interface
//******************************************************************
// RAM organization: 33792 words * 18 bits (last address = 83FFh)
scope_mem_sp buffer_ram(
	.address		(ram_address[15:0]),
	.inclock		(PLL0_sampleCLK),
	.outclock	(~PLL0_sampleCLK),
	.data			(record[17:0]),
	.wren			(ram_wr),
	.q				(ram_out[17:0])
);

//******************************************************************
// Host interface
//******************************************************************
/* port 30 = low byte of RAM address to read (rw)
	port 31 = (bits 5..0) high 6 bits of RAM address (rw)
	port 32 = sampled data from the RAM buffer address designated by ports 30 and 31 (r)
	port 33 = control line state (bits 3..0) (r)
	port 34 = (bit 7) sampling start (rw), (bit 6) sampling end (r), duration of the _previous_ state (bits bits 5..0)
	port 38 = sampling time counter (low)
	port 39 = sampling time counter
	port 3a = sampling time counter
	port 3b = sampling time counter (high)

 Algorithm: reset sampling_start to ensure proper initialization of the scope.
 Set sampling_start, operation starts. Read sampling_end until it is 1. Then read
 sampling time counter. Then reset sampling_start, and read RAM buffer.
*/

wire port30 = (BA[7:0] == 8'h30) & ~BIORQ_N & BM1_N;
wire port30_write = port30 & ~BWR_N;
wire port30_read = port30 & ~BRD_N;
wire port31 = (BA[7:0] == 8'h31) & ~BIORQ_N & BM1_N;
wire port31_write = port31 & ~BWR_N;
wire port31_read = port31 & ~BRD_N;

wire port32_read = (BA[7:0] == 8'h32) & ~BIORQ_N & BM1_N & ~BRD_N;
wire port33_read = (BA[7:0] == 8'h33) & ~BIORQ_N & BM1_N & ~BRD_N;
wire port34 = (BA[7:0] == 8'h34) & ~BIORQ_N & BM1_N;
wire port34_write = port34 & ~BWR_N;
wire port34_read = port34 & ~BRD_N;

wire port38_read = (BA[7:0] == 8'h38) & ~BIORQ_N & BM1_N & ~BRD_N;
wire port39_read = (BA[7:0] == 8'h39) & ~BIORQ_N & BM1_N & ~BRD_N;
wire port3a_read = (BA[7:0] == 8'h3a) & ~BIORQ_N & BM1_N & ~BRD_N;
wire port3b_read = (BA[7:0] == 8'h3b) & ~BIORQ_N & BM1_N & ~BRD_N;

// host data bus output data control
assign BD[7:0] =	port30_read ? host_read_address[7:0] :
						port31_read ? host_read_address[15:8] :
						port32_read ? ram_out[7:0] :													// data (8b)
						port33_read ? {{4{1'b0}}, ram_out[11:8]} :								// control lines (4b)
						port34_read ? {sampling_start, sampling_end, ram_out[17:12]} :		// duration (6b)
						port38_read ? cur_time[7:0] :
						port39_read ? cur_time[15:8] :
						port3a_read ? cur_time[23:16] :
						port3b_read ? cur_time[31:24] :						
						{8{1'bz}};

// host data bus output direction control
assign m_dio = port30_read | port31_read | port32_read | port33_read | port34_read |
					port38_read | port39_read | port3a_read | port3b_read;

always @(posedge PLL0_sampleCLK or negedge BRESET_N) begin
	if(!BRESET_N) begin
		sampling_start <= 1'b0;
		bclock <= 1'b0;
	end
	else begin
		if(I_BCLOCK && !bclock) begin										// latch data from hsot only on posedge of system clock
			if(port30_write) host_read_address[7:0] <= BD[7:0];	// low byte of the address
			if(port31_write) host_read_address[15:8] <= BD[7:0];	// high byte of the address (6 bits effective only)
			if(port34_write) sampling_start <= BD[7];					// sampling start request bit
		end
		bclock <= I_BCLOCK;
	end
end

endmodule
/*********************************************************************************
 *			SCOPE_PLUGIN.V																				*
 ********************************************************************************/