/***************************************************************************************
 *																													*
 *	SCC_M9K.V:	Konami SCC sound chip function implementation for GR8BUS master board,	*
 *															based on Altera Cyclone III devices			*
 *															(uses M9K block and embedded multiplier)	*
 *																					http://www.gr8bit.ru		*
 *	Copyleft (cl) 2012 by Eugeny Brychkov																*
 *																													*
 *									License and T&C															*
 *																													*
 * You can use and modify this software's source and executable, given two					*
 * conditions are met: (a) you keep this Copyleft statement in the source code,			*
 * and you do not distribute software and resulting executables on the commercial		*
 * basis and for the purposes, not related to GR8BIT platform and/or MSX standard,		*
 * without written permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)			*
 *																													*
 ***************************************************************************************
 *			File name: SCC_M9K.V																				*
 **************************************************************************************/

 // Released:	December 02, 2012
 // NOTICE:	Quartus's TimeQuest Timing Analyzer may issue critical warnings for the code,
 // 			but nevertheless code works properly

module scc_m9k(
	input wire i_clock_low,				// low speed clock, used for waveform timing (expected standard 3.58MHz frequency)
	input wire i_clock_high,			// high speed clock, used for RAM access and mixer synthesis (at least 16 times faster than low speed clock)
	input wire i_reset_n,				// system reset signal (used to turn off all the channels)
	
	input wire [15:0] i_addr,			// higher level module address bus (GR8BUS)
	input wire [7:0] i_data,			// higher level module input data bus (GR8BUS)
	input wire i_devselect,				// device (module) select (derived from /SLTx signal)
	input wire i_rd_n,					// device read, negative (GR8BUS/Z80)
	input wire i_wr_n,					// device write, negative (GR8BUS/Z80)

	output reg [5:0] DAC,				// DAC output, 6-bit

	output reg [7:0] o_data,			// SCC module output data to the higher level module data bus
	output wire o_direction				// SCC module data bus direction flag (=1 means data from SCC to CPU)
);

// SCC register space definition
reg [11:0] reg_tone[0:4];				// SCC tone registers (5), 12-bit
reg [3:0] reg_volume[10:14];			// SCC volume registers (5), 4-bit. Indexing from 10 is very important!
reg [4:0] reg_channel_enable;			// Channel activity flags
reg [7:0] reg_deformation;				// deformation register (not used in this implementation)

// wave RAM definition (uses one M9K block)
reg [6:0] wave_mem_addr;				// latched memory address
reg [7:0] wave_mem_di;					// latched input data
reg wave_mem_we;							// latched read or write signal
wire [7:0] wave_mem_do;					// RAM data output wires

// waveform RAM definition (128 bytes, uses M9K blocks)
scc_ram mem_wave(
	.address	(wave_mem_addr[6:0]),
	.clock	(i_clock_high),			// high speed clock is used, RAM op starts at rising edge, result is on following falling edge
	.data		(wave_mem_di[7:0]),
	.wren		(wave_mem_we),
	.q			(wave_mem_do[7:0])
);

// control signals derived from signals supplied by higher level module
wire SCC_select = (i_addr[15:11] == 5'b10011) & i_devselect && (konami_mem_page3[5:0] == 6'b111111);	// activates for SCC addresses and mapper page 03fh
wire wave_table_access = SCC_select & ~i_addr[7];				// activates on waveform address table area access
wire wave_table_read = wave_table_access & !i_rd_n;			// activates for reads from the waveform table
wire wave_table_write = wave_table_access & !i_wr_n;			// activates on writes to the waveform table
wire control_reg_write = SCC_select & !i_wr_n & i_addr[7];	// activates for all writes into SCC address space
wire SCC_page_write = (i_addr[15:11] == 5'b10010) & i_devselect & !i_wr_n;		// activates on Konami memory mapper write, bank 3

// data bus direction for higher level module
assign o_direction = wave_table_read;

// Konami memory mapper register (5 meaningful bits only)
reg [5:0] konami_mem_page3;		// current page of the Konami memory mapper (for page 3 only)

//=======================================================================================================================
// Waveform RAM data I/O access and mixing section
//=======================================================================================================================
reg [2:0] ram_stage;					// stage of the RAM access/mixing process
reg [15:0] mixer_value;				// final value of the mixer to be fed up to the DAC
reg [3:0] waveform_volume;			// latched current channel volume for multiplication
reg [15:0] waveform_to_add;		// result of the multiplication. Please see the multiplier section for description
											// of the multiplication and formation of mixed waveform value

// --- Combined host RAM access and waveform mixing procedure ---
/*	As soon as low speed clock (3.58Mhz) is identified to be high during falling edge of the high speed clock,
	the staged process is started. Every stage lasts for one period of high speed clock - RAM inputs are set up
	on the falling edge of the high freq clock, and clocked in on rising edge. On the next falling edge data is
	is sampled off the RAM (and in parallel new input values are set up).
	- Stage 000 is RAM r/w command execution basing on address/data/wren from the higher level module.
	  If higher level module (GR8BUS/Z80) did not request any access, dummy read is performed;
	- Stage 001 is setting up input address/data for waveform data read for channel 1. Data read off the RAM
	  is used in the following always block for mixing;
	- Stages 010-101 are same reads for channels 2-5. Data is sampled in following always block on negative edge of
	  high speed clock;
	- Stage 110 is the last stage when channel 5 data is sampled. It is required for peoper data sampling process;
	- Stage 111 is just waiting for low speed clock getting low.	*/
always @(negedge i_clock_high or negedge i_reset_n) begin
	if(!i_reset_n) begin					// reset condition: initialize variables for this block
		ram_stage[2:0] <= 3'b111;		// jump into process of waiting of low speed clock come inactive
		wave_mem_we <= 1'b0;				// next RAM access operation is read (default)
	end
	else
		case(ram_stage[2:0])
			3'b000: begin					// wait for low speed clock become high, and start processing CPU access cycle
				if(i_clock_low) begin	// low speed clock becomes high level, if there's CPU access cycle its control lines will signal so
					wave_mem_addr[6:0] <= i_addr[6:0];				// set RAM address to current CPU address
					wave_mem_di[7:0] <= i_data[7:0];					// set RAM input data bus to current CPU data bus
					if(wave_table_write)	wave_mem_we <= 1'b1;		// set RAM write line 
					ram_stage[2:0] <= 3'b001;
				end
			end
			3'b001: begin												// time slot for sampling CPU data and start reading first waveform
				wave_mem_we <= 1'b0;									// important: reset RAM data write bit if it was set previously
				wave_mem_addr[6:0] <= ch1index[6:0];			// channel 1 address
				ram_stage[2:0] <= 3'b010;
			end
			3'b010: begin												// time slot for start reading second waveform
				wave_mem_addr[6:0] <= ch2index[6:0];			// channel 2 address
				ram_stage[2:0] <= 3'b011;
			end
			3'b011: begin												// time slot for start reading third waveform
				wave_mem_addr[6:0] <= ch3index[6:0];			// channel 3 address
				ram_stage[2:0] <= 3'b100;
			end
			3'b100: begin												// time slot for start reading fourth waveform
				wave_mem_addr[6:0] <= ch4index[6:0];			// channel 4 address
				ram_stage[2:0] <= 3'b101;
			end
			3'b101: begin												// time slot for start reading fifth waveform
				wave_mem_addr[6:0] <= ch5index[6:0];			// channel 5 address
				ram_stage[2:0] <= 3'b110;
			end
			3'b110: ram_stage[2:0] <= 3'b111;					// last data byte (waveform for channel 5) will be latched
																			// at this negative edge of high speed clock
			3'b111:
				if(!i_clock_low)										// wait in 3'b111 stage until low speed clock deactivates,
					ram_stage[2:0] <= 3'b000;						// and go to stage 3'b000
		endcase
end

// --- RAM data access and final waveform mixing ---
/*	RAM data becomes available on the following negative edge of its input clock (high speed clock).
	Data is latched in the order described for always block above. First data is written or read for host access.
	Following 5 cycles are reads of specific bytes off the waveform area, immediate multiplication with the respective
	channel volume (or *0 if channel is disabled) and result at every cycle is added up to the mixer value.
	Important assumption: we assume (and it is acknowledged by tests) that on the following negedge of high speed clock not only data
	from RAM is ready, but it also allowed enough time for its multiplication with volume.	*/
always @(negedge i_clock_high)
	case(ram_stage[2:0])
		3'b001: begin
			if(!wave_mem_we && wave_table_read)	o_data[7:0] <= wave_mem_do[7:0];				// RAM data read for CPU
			mixer_value[15:0] <= 1'b0;																		// clearing mixer value
			waveform_volume[3:0] <= reg_channel_enable[0] ? reg_volume[10][3:0] : 1'b0;	// setting current volume for multiplication (ch 1)
		end
		3'b010: begin
			mixer_value[15:0] <= mixer_value[15:0] + waveform_to_add[15:0];					// at this stage data byte and its multiplication is ready for channel 1
			waveform_volume[3:0] <= reg_channel_enable[1] ? reg_volume[11][3:0] : 1'b0; 	// setting current volume for multiplication (ch 2)
		end
		3'b011: begin
			mixer_value[15:0] <= mixer_value[15:0] + waveform_to_add[15:0];					// at this stage data and its multiplication is ready for channel 2
			waveform_volume[3:0] <= reg_channel_enable[2] ? reg_volume[12][3:0] : 1'b0;	// setting current volume for multiplication (ch 3)
		end
		3'b100: begin
			mixer_value[15:0] <= mixer_value[15:0] + waveform_to_add[15:0];					// at this stage data and its multiplication is ready for channel 3
			waveform_volume[3:0] <= reg_channel_enable[3] ? reg_volume[13][3:0] : 1'b0;	// setting current volume for multiplication (ch 4)
		end
		3'b101: begin
			mixer_value[15:0] <= mixer_value[15:0] + waveform_to_add[15:0];					// at this stage data and its multiplication is ready for channel 4
			waveform_volume[3:0] <= reg_channel_enable[4] ? reg_volume[14][3:0] : 1'b0;	// setting current volume for multiplication (ch 5)
		end
		3'b110: begin
			mixer_value[15:0] <= mixer_value[15:0] + waveform_to_add[15:0];					// at this stage data and its multiplication is ready for channel 5
		end
		// at the end final waveform value in mixer_value is ready to be clocked into the DAC on the following posedge of _low speed_ clock
	endcase
	
//=======================================================================================================================
// Higher level module (GR8BUS/Z80) access section
//=======================================================================================================================
/*	Host read/write condition is synchronized with positive edge of the low speed clock (3.58MHz, Z80 CPU clock)
	(for more information about Z80 timing, please refer to Z80 CPU user's manual, www.zilog.com/docs/z80/um0080.pdf)
	(for SCC access information please refer to the Wikipedia article http://en.wikipedia.org/wiki/Konami_SCC).	*/
always @(posedge i_clock_low or negedge i_reset_n) begin
	if(!i_reset_n)
		reg_channel_enable[4:0] <= 5'b00000;			// on reset disable all channels, other regs do not care
	else begin
		// SCC control registers data write
		if(control_reg_write)								// only control register writes, waveform area access is served in
																	// "waveform data I/O access section"
			if(i_addr[6:5] == 2'b00) begin
				if(i_addr[3:1] < 3'b101) begin				// tone registers? (addresses 0000-1010 (cells 0 to 9))
					if(i_addr[0])	reg_tone[i_addr[3:1]][11:8] <= i_data[3:0];	// yes, MSByte
					else	reg_tone[i_addr[3:1]][7:0] <= i_data[7:0];				// yes, LSByte
				end	else
						if(i_addr[3:0] == 4'b1111)								// channel enable register? (cell 15)
							reg_channel_enable[4:0] <= i_data[4:0];		// yes, write to channel enable register
						else	reg_volume[i_addr[3:0]] <= i_data[3:0];	// otherwise volume registers (important: note "reg_volume"
																						// definition index starts with value 10)
			end
				else	if(i_addr[6:4] == 2'b11)	reg_deformation[7:0] <= i_data[7:0];	// otherwise deformation register (not used)
		// Konami memory mapper page definition (bank 3)
		if(SCC_page_write)	konami_mem_page3[5:0] <= i_data[5:0];
	end
end

//=======================================================================================================================
// Tone generation and pointers into the waveforms
//=======================================================================================================================
// Current tone counters for all 5 channels, 12-bit each
reg [11:0] ch1tone;
reg [11:0] ch2tone;
reg [11:0] ch3tone;
reg [11:0] ch4tone;
reg [11:0] ch5tone;
// waveform arrays pointers, 5-bit (32 bytes in each waveform array)
reg [4:0] ch1pos;
reg [4:0] ch2pos;
reg [4:0] ch3pos;
reg [4:0] ch4pos;				// remember that channels 4 and 5 share waveform area, but have different control registers
reg [4:0] ch5pos;				// but have different control registers and act independently

/*	This block processes tone counters for all 5 channels, shifting +1 pointer for the waveform byte
	of the channel which tone countdown counter is reached 0. Unlike PSG which tone value of 0
	means 4096 cycles (very low frequency), for SCC tone value of 0 means highest frequency (i.e
	increasing channel waveform pointer every low speed clock cycle (checked on the real SCC chip).	*/
always @(posedge i_clock_low) begin
	// channel 1 tone generation
	if(ch1tone[11:0] == {12{1'b0}}) begin					// channel 1 countdown reached: next sample
		ch1tone[11:0] <= reg_tone[0][11:0];					// reset tone counter
		ch1pos[4:0] <= ch1pos[4:0] + 1'b1;					// increase position in the waveform area (circular, w/o overflow)
	end	else	ch1tone[11:0] <= ch1tone[11:0] - 1'b1;	// otherwise decrease counter for channel 1
	
	// channel 2 tone generation
	if(ch2tone[11:0] == {12{1'b0}}) begin					// channel 2 countdown reached: next sample
		ch2tone[11:0] <= reg_tone[1][11:0];
		ch2pos[4:0] <= ch2pos[4:0] + 1'b1;
	end	else	ch2tone[11:0] <= ch2tone[11:0] - 1'b1;	// otherwise decrease counter for channel 2

	// channel 3 tone generation
	if(ch3tone[11:0] == {12{1'b0}}) begin					// channel 3 countdown reached: next sample
		ch3tone[11:0] <= reg_tone[2][11:0];
		ch3pos[4:0] <= ch3pos[4:0] + 1'b1;
	end	else	ch3tone[11:0] <= ch3tone[11:0] - 1'b1;	// otherwise decrease counter for channel 3

	// channel 4 tone generation
	if(ch4tone[11:0] == {12{1'b0}}) begin					// channel 4 countdown reached: next sample
		ch4tone[11:0] <= reg_tone[3][11:0];
		ch4pos[4:0] <= ch4pos[4:0] + 1'b1;
	end	else	ch4tone[11:0] <= ch4tone[11:0] - 1'b1;	// otherwise decrease counter for channel 4

	// channel 5 tone generation
	if(ch5tone[11:0] == {12{1'b0}}) begin					// channel 5 countdown reached: next sample
		ch5tone[11:0] <= reg_tone[4][11:0];
		ch5pos[4:0] <= ch5pos[4:0] + 1'b1;
	end	else	ch5tone[11:0] <= ch5tone[11:0] - 1'b1;	// otherwise decrease counter for channel 5
end

//=======================================================================================================================
// Multiplication section: waveform by its volume
//=======================================================================================================================
/*	Definition of the positions of the current waveform byte in the whole 128-byte memory space, top 2 bits
	define the "bank" for the waveform, and next 5 bits define position within specified channel's 32-bit area. */
wire [6:0] ch1index = { 2'b00, ch1pos[4:0]};
wire [6:0] ch2index = { 2'b01, ch2pos[4:0]};
wire [6:0] ch3index = { 2'b10, ch3pos[4:0]};
wire [6:0] ch4index = { 2'b11, ch4pos[4:0]};
wire [6:0] ch5index = { 2'b11, ch5pos[4:0]};

/*	Signed multiplication of current waveform by current volume is performed "on-the-fly" just after valid data is read
	from the RAM area in the mixing section, and then result on the negedge of high speed clock is added to the mixer value.
	Input lines are 8-bit each, signed values, so at the output we have 16-bit signed value. Please note "datab" argument,
	which is "signed positive" with 0s added at the MSB(s). Shifting "waveform_volume" value right within 8-bit value
	will increase volume of the cumulative sound, but may cause final mixer value overflow and corrupt sound. Having
	"waveform_volume" too "left" in 8-bit value will decrease volume, and may cause poor performance of the sound through
	6-bit DAC. In this implementation 4-bit unsigned "waveform_volume" value appears preceded by two 0s and followed by two 0s,
	seems not causing final mixer value overflow in known conditions.	*/
lpmmult waveform_by_volume(									// uses single embedded multiplier 9-bit element
	.dataa (wave_mem_do[7:0]),									// data byte from the waveform RAM
	.datab ({ 2'b00, waveform_volume[3:0], 2'b00 }),	// current waveform volume, signed positive, *4
	.result (waveform_to_add[15:0])							// wires to the adder
);

//=======================================================================================================================
// Output to the DAC
//=======================================================================================================================
/*	Strobing mixed value to DAC device - on the negative edge of the reference clock, when calculation of
	the "mixer_value" had been completed. For implementation to work properly it is required to have high speed
	clock at least 16 times faster than low speed clock (i.e. >= 57.28Mhz). When developing this code, we used
	high speed clock equal to 85.92Mhz.	*/
always @(negedge i_clock_low) begin
	DAC[4:0] <= mixer_value[14:10];
	DAC[5] <= ~mixer_value[15];				// centering waveform in the middle of the DAC voltage scale
end

endmodule

/*********************************************************************************
 *			END OF: SCC_M9K.V																			*
 ********************************************************************************/
