/***************************************************************************************
 *																													*
 *	LAN_SPI_MODULE.V:		Module to access LAN chip WIZnet W5100 of the GR8BUS master		*
 *								board built on Altera Cyclone III device								*
 *																						http://www.gr8bit.ru	*
 *	Copyleft (cl) 2012 by Eugeny Brychkov																*
 *																													*
 *									License and T&C															*
 *																													*
 * You can use and modify this software's source and executable, given two					*
 * conditions are met: (a) you keep this Copyleft statement in the source code,			*
 * and you do not distribute software and resulting executables on the commercial		*
 * basis and for the purposes, not related to GR8BIT platform and/or MSX standard,		*
 * without written permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)			*
 *																													*
 ***************************************************************************************
 *			File name: LAN_SPI_MODULE.V																	*
 **************************************************************************************/

/* Release date: Dec 20, 2012

	Techical information:
		- MISO and MOSI pins of the FPGA device may need weak pull-ups to be activated
		- LAN module has its own SCLK signal to the SPI bus, and works in SPI mode 0
		- host port 0 is used for storing low LAN address, port 1 for storing high LAN address
		- MSB bit of address register is used as "address auto-incdement" bit if it's 1 (unimplemented yet)
		- host port 2 is used for data I/O with W5100. Access takes 34*2 cycles of "clk_2xSPI_clock",
		  and wait signal is generated during W5100 SPI exchange
		- module is designed to work with SPI bus arbiter, but may work in standalone configuration
*/

module LAN_SPI_module(
	// clocks and resets
	input wire clk_2xSPI_clock,				// clock signal to derive SPI clock from, 2x value
	input wire reset_n,							// reset signal
	input wire clk_sysbus_clock,				// system bus clock
	// system bus interface
	input wire i_LAN_cs,							// LAN module - port access active
	input wire [1:0] i_LAN_portnum,			// LAN module - port number (00 = addrlow, 01 = addrhigh, 10=data)
	input wire i_rd_n,							// bus read signal
	input wire i_wr_n,							// bus write signal
	input wire [7:0] i_data,					// input data bus
	
	output wire o_direction,					// data output direction to the system data bus
	output wire [7:0] o_data,					// output data bus (register)
	output wire o_wait_signal_n,				// wait signal while SPI exchange is in progress

	// SPI bus and SPI arbiter interface
	output reg o_LAN_SPI_request,				// SPI bus request (for arbiter)
	input wire i_LAN_timeslot,					// becomes active when SPI arbiter gives activity timeslot to LAN module
	output reg o_LAN_SPI_SCLK,					// SCLK from/for LAN SPI exchange (register)
	output reg o_LAN_SPI_CS_n,					// CS signal from master to LAN chip (register)
	output reg o_LAN_SPI_MOSI,					// MOSI signal from master to LAN chip (register)
	input wire i_LAN_SPI_MISO					// MISO signal from LAN chip to master (wire)
);


//-------------------------------------------------------------------------------------
// 1. LAN module declarations
//-------------------------------------------------------------------------------------
reg [15:0] LAN_address;			// LAN address register for access
reg [6:0] LAN_SPI_stage;		// stage of the SPI access process
										// bit 0 = trigger for polarity of SCLK (inverted)
										// bit 6 = flag of the end of the sequence
										// bits 5:1 = current bit number +1
wire [6:0] LAN_sequence_bit_num = LAN_SPI_stage[6:0] - 1'b1;		// used as index for the current bit number in registers
reg [7:0] LAN_o_data;			// SPI read data latched and presented to CPU
reg [7:0] LAN_i_data;			// data from CPU to shift into the SPI channel

//-------------------------------------------------------------------------------------
// 2. Wire states
//-------------------------------------------------------------------------------------
// LAN module port 0 (address LOW) read/write signals
wire LAN_port0 = i_LAN_cs & (i_LAN_portnum[1:0] == 2'b00);		// port 0 active
wire LAN_port0_read = LAN_port0 & ~i_rd_n;							// port 0 read
reg LAN_port0_read_detect;
wire LAN_port0_write = LAN_port0 & ~i_wr_n;							// port 0 write
reg LAN_port0_write_detect;
// LAN module port 1 (address HIGH) read/write signals
wire LAN_port1 = i_LAN_cs & (i_LAN_portnum[1:0] == 2'b01);
wire LAN_port1_read = LAN_port1 & ~i_rd_n;
reg LAN_port1_read_detect;
wire LAN_port1_write = LAN_port1 & ~i_wr_n;
reg LAN_port1_write_detect;
// LAN module port 2 (data exchange with W5100) read/write signals
wire LAN_port2 = i_LAN_cs & (i_LAN_portnum[1:0] == 2'b10);
wire LAN_port2_read = LAN_port2 & ~i_rd_n;
reg LAN_port2_read_detect;
wire LAN_port2_write = LAN_port2 & ~i_wr_n;
reg LAN_port2_write_detect;
wire LAN_port2_write_posedge = LAN_port2_write && !LAN_port2_write_detect;
wire LAN_port2_read_posedge = LAN_port2_read && !LAN_port2_read_detect;


wire LAN_SPI_sequence_end = (LAN_SPI_stage[6:0] == 7'b1111111);	// becomes active when SPI exchange sequence finishes
wire miso_sample_time = LAN_port2_read & (LAN_SPI_stage[6:4] == 3'b000) && (~LAN_SPI_stage[0]);	// time when data should be sampled
																																// from the MISO line

assign o_direction = LAN_port0_read | LAN_port1_read | LAN_port2_read;	// system data bus direction signal
assign o_data = LAN_port0_read ? LAN_address[7:0] : LAN_port1_read ? LAN_address[15:8] : LAN_port2_read ? LAN_o_data[7:0] : {8{1'bz}};
assign o_wait_signal_n = LAN_port2 ? (~o_LAN_SPI_request) : 1'b1;			// on LAN chip access, activate
																								// system WAIT signal until LAN chip access is complete
wire [15:0] LAN_address_MOSI = { 1'b0, LAN_address[14:0] };					// address to shift to LAN chip through SPI channel, MSB is 0

/*-----------------------------------------------------------------------------------
  3. Write to the LAN address register
  -----------------------------------------------------------------------------------*/
// On the rising edge of the system bus clock (Z80 CPUCLK) checking for positive
// edge of port0 and port1 write signals, and sample data into respective part of register
always @(posedge clk_sysbus_clock or negedge reset_n) begin
	if(!reset_n) begin
		LAN_address[15:0] <= 16'b0;
		LAN_port0_write_detect <= 1'b0;
		LAN_port1_write_detect <= 1'b0;
	end
	else begin
		if(LAN_port0_write && !LAN_port0_write_detect)	// positive edge of the CS is detected, port 0
			LAN_address[7:0] <= i_data[7:0];					// LAN address low
		if(LAN_port1_write && !LAN_port1_write_detect)	// positive edge of the CS is detected, port 1
			LAN_address[15:8] <= i_data[7:0];					// LAN address high

//		if(!LAN_address[15])						// LAN address auto-increment bit is 0?
//			LAN_address[14:0] <= LAN_address[14:0] + 1'b1;	// yes, increase 15 bits of the LAN address

		// preserving current state of the CS signals
		LAN_port0_write_detect <= LAN_port0_write;
		LAN_port1_write_detect <= LAN_port1_write;
	end
end

/*-----------------------------------------------------------------------------------
  4. SPI requence request, initiation and completion
  -----------------------------------------------------------------------------------*/
always @(posedge clk_2xSPI_clock or negedge reset_n) begin
	if(!reset_n) begin
		LAN_port2_write_detect <= 1'b0;
		LAN_port2_read_detect <= 1'b0;
		o_LAN_SPI_request <= 1'b0;						// reset request signal on reset
	end
	else	begin
		// detect [positive edge] of the LAN dataport (device) access request - activation
		if(LAN_port2_write_posedge)					// positive edge of the CS is detected, port 2
			LAN_i_data[7:0] <= i_data[7:0];			// latch data to write to LAN chip in writing sequence in always block #5.
		
		if(LAN_port2_write_posedge || LAN_port2_read_posedge)	// posedge of port 2 read or write signals?
			o_LAN_SPI_request <= 1'b1;									// activate SPI bus request signal
		// if SPI command for LAN device reached its end, reset SPI request signal
		if(LAN_SPI_sequence_end)
			o_LAN_SPI_request <= 1'b0;					// reset LAN SPI request at the end of sequence

		// update current status of the LAN data port select signal
		LAN_port2_write_detect <= LAN_port2_write;
		LAN_port2_read_detect <= LAN_port2_read;
	end
end


/*-----------------------------------------------------------------------------------
  5. Writing sequence to the LAN chip
  -----------------------------------------------------------------------------------*/
always @(posedge clk_2xSPI_clock) begin
	if(!reset_n || !i_LAN_timeslot) begin			// initialization happens on reset and if
																// LAN module is not give timeslot by arbiter
		LAN_SPI_stage[6:0] <= 7'b1000000;			// initialize the first stage count
		o_LAN_SPI_CS_n <= 1'b1;							// chip deselected
		o_LAN_SPI_SCLK <= 1'b0;							// SCLK ls low (important! - SPI mode 0)
		o_LAN_SPI_MOSI <= 1'b1;
	end
	else begin
		if(i_LAN_timeslot) begin						// within arbitrated timeslot?
			if(LAN_sequence_bit_num[0]) begin		// is it time to put data into the MOSI (negedge of SCLK)
				o_LAN_SPI_CS_n <= 1'b0;					// LAN device chip select - activate
				case(LAN_sequence_bit_num[6:4])		// stage by byte
					3'b011: begin							// first byte of the SPI sequence, command
						if(LAN_port2_read) o_LAN_SPI_MOSI <= ~LAN_sequence_bit_num[3];		// 00001111, read operation
						else	o_LAN_SPI_MOSI <= LAN_sequence_bit_num[3];						// 11110000, write operation
					end
					3'b010,
					3'b001:									// second and third bytes of the SPI sequence, address
						o_LAN_SPI_MOSI <= LAN_address_MOSI[ { ~LAN_sequence_bit_num[4] , LAN_sequence_bit_num[3:1] } ];	// shift LAN address from #1111 to #0000
					3'b000:									// last byte: data byte
						o_LAN_SPI_MOSI <= LAN_i_data[LAN_sequence_bit_num[3:1]];
				endcase
			end
			if(LAN_SPI_sequence_end)
				o_LAN_SPI_CS_n <= 1'b1;					// LAN device chip select - sequence ended, deselected
			if(!LAN_SPI_stage[6])						// disables alternating of the SCLK signal at the first data setup
																// and at the end of the sequence
				o_LAN_SPI_SCLK <= ~o_LAN_SPI_SCLK;	// alternate SCLK clock
			// decrease sequence number
			if(!LAN_SPI_sequence_end) LAN_SPI_stage[6:0] <= LAN_SPI_stage[6:0] - 1'b1;
		end
	end
end

// following always block reads byte from the LAN chip using MISO lines
always @(posedge clk_2xSPI_clock) begin
	if(miso_sample_time)												// time to sample respective bit?
		LAN_o_data[LAN_SPI_stage[3:1]] <= i_LAN_SPI_MISO;	// sample bit identified by number in [3:1]
end

endmodule

/*********************************************************************************
 *			LAN_SPI_MODULE.V																			*
 ********************************************************************************/
