/***************************************************************************************
 *																													*
 *	SPI_ARBITER.V:			Module to arbiter SPI access through the GR8BUS master			*
 *								board SPI channel built on Altera Cyclone III device				*
 *																						http://www.gr8bit.ru	*
 *	Copyleft (cl) 2012 by Eugeny Brychkov																*
 *																													*
 *									License and T&C															*
 *																													*
 * You can use and modify this software's source and executable, given two					*
 * conditions are met: (a) you keep this Copyleft statement in the source code,			*
 * and you do not distribute software and resulting executables on the commercial		*
 * basis and for the purposes, not related to GR8BIT platform and/or MSX standard,		*
 * without written permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)			*
 *																													*
 ***************************************************************************************
 *			File name: SPI_ARBITER.V																		*
 **************************************************************************************/

/* Release date: Dec 20, 2012

	Techical information:
	- all devices using arbiter should work in SPI mode 0 -> SCLK signal idle low
	- one request -> one transaction. All other devices are on hold while transaction is in progress
	- uses round-robin time-division method for simplicity. With 3 configured devices delays between
	  two consecutive requests of the device (given other devices do not request service) is 3 arbiter clock cycles
	- each SPI device can use its own clock source of required frequency
	- How it works:
	  * module asserts its SPIrequest output signal (active high), and watches for its timeslot signal to become active (high)
	  * as soon as timeslot signal becomes high, device prepares data for first transfer, and starts its SCLK with
	    device-specific clocking frequency
	  * at the end of SPI exchange, device deactivates its SCLK signal (low), and then deactivates SPIrequest line,
	    and on next positive edge of arbiter's clock arbiter switches to next device. It is advisable for SPI module to
	    complete exchange with its hardware properly - i.e. put its SCLK low and deassert device's CS signal before
	    deactivating its SPIrequest signal
*/

module SPI_arbiter(
	// arbiter clock, the speed with which arbiter checks device requests. This clock can be as fast as FPGA chip is able
	input wire arbiter_clock,

	// reset signal to initialize arbiter state
	input wire reset_n,

	// LAN module wires
	input wire i_LAN_SPIrequest,	// SPI bus request signal from the LAN control module
	output reg o_LAN_timeslot,		// Channel free signal (active high) indicating that LAN
											//   module can start its LAN_cs/MOSI/MISO sequence
	input wire i_LAN_clock,			// LAN module SPI clock
	input wire i_LAN_cs,				// LAN module chip select for LAN device when access is granted
	input wire i_LAN_mosi,			// LAN module MOSI signal when access is granted
	output wire o_LAN_miso,			// LAN module MISO signal when access if granted

	// SD-card module wires
	input wire i_SD_SPIrequest,	// SPI bus request signal from the SD control module
	output reg o_SD_timeslot,		// Channel free signal (active high) indicating that SD-card
											//   module can start its SD_cs/MOSI/MISO sequence
	input wire i_SD_clock,			// SD-card module SPI clock
	input wire i_SD_cs,				// SD-card module chip select for SD card device when access is granted
	input wire i_SD_mosi,			// SD-card module MOSI signal when access is granted
	output wire o_SD_miso,			// SD-card module MISO signal when access if granted

	// USB module wires
	input wire i_USB_SPIrequest,	// SPI bus request signal from the USB control module
	output reg o_USB_timeslot,		// Channel free signal (active high) indicating that USB
											//   module can start its USB_cs/MOSI/MISO sequence
	input wire i_USB_clock,			// USB module SPI clock
	input wire i_USB_cs,				// USB module chip select for USB device when access is granted
	input wire i_USB_mosi,			// USB module MOSI signal when access is granted
	output wire o_USB_miso,			// USB module MISO signal when access if granted

	// SPI bus pins
	output wire p_LANCS_n,			// SPI channel LAN device select 
	output wire p_SDCS_n,			// SPI channel SD-card device select 
	output wire p_USBCS_n,			// SPI channel USB device select 
	output wire p_sclk,				// SPI channel serial clock signal
	output wire p_mosi,				// SPI channel master-out-slave-in signal 
	input wire p_miso					// SPI channel master-in-slave-out signal 
);

//-------------------------------------------------------------------------------------
// 1. Variables
//-------------------------------------------------------------------------------------
reg [1:0] round_robin;				// counter for round robin tiem division for the module's access request
											// 00 = LAN, 01 = SD, 10 = USB, 11 = illegal

//-------------------------------------------------------------------------------------
// 2. Wire states
//-------------------------------------------------------------------------------------
// multiplexing SPI signals
// modules -> device
assign p_mosi = o_LAN_timeslot ? i_LAN_mosi : o_SD_timeslot ? i_SD_mosi : o_USB_timeslot ? i_USB_mosi : 1'b1;
assign p_sclk = o_LAN_timeslot ? i_LAN_clock : o_SD_timeslot ? i_SD_clock : o_USB_timeslot ? i_USB_clock : 1'b0;	// Notice: SCLK idle LOW
assign p_LANCS_n = o_LAN_timeslot ? i_LAN_cs : 1'b1;
assign p_SDCS_n = o_SD_timeslot ? i_SD_cs : 1'b1;
assign p_USBCS_n = o_USB_timeslot ? i_USB_cs : 1'b1;
// devices -> modules
assign o_LAN_miso = o_LAN_timeslot ? p_miso : 1'b1;
assign o_SD_miso = o_SD_timeslot ? p_miso : 1'b1;
assign o_USB_miso = o_USB_timeslot ? p_miso : 1'b1;

//-------------------------------------------------------------------------------------
// 3. Performing multiplex operation
//-------------------------------------------------------------------------------------
always @(posedge arbiter_clock or negedge reset_n) begin
	if(!reset_n) begin
		round_robin[1:0] <= 2'b0;		// start with LAN module (initial value actually does not matter)
		o_LAN_timeslot <= 1'b0;			// none module have access to SPI channel
		o_SD_timeslot <= 1'b0;
		o_USB_timeslot <= 1'b0;
	end
	else begin
		case(round_robin[1:0])
			2'b00: begin										// LAN chip access timeframe
				o_LAN_timeslot <= i_LAN_SPIrequest;		// LAN is given a timeslot if it requested service
				if(!i_LAN_SPIrequest)						// however if it did not request, skip to next device
					if(round_robin[1:0] == 2'b10) round_robin[1:0] <= 2'b00;
					else	round_robin[1:0] <= round_robin[1:0] + 1'b1;
			end
			2'b01: begin
				o_SD_timeslot <= i_SD_SPIrequest;		// same for SD-card
				if(!i_SD_SPIrequest)
					if(round_robin[1:0] == 2'b10) round_robin[1:0] <= 2'b00;
					else	round_robin[1:0] <= round_robin[1:0] + 1'b1;
			end
			2'b10: begin
				o_USB_timeslot <= i_USB_SPIrequest;		// same for USB chip
				if(!i_USB_SPIrequest)
					if(round_robin[1:0] == 2'b10) round_robin[1:0] <= 2'b00;
					else	round_robin[1:0] <= round_robin[1:0] + 1'b1;
			end
		endcase
	end
end

endmodule

/*********************************************************************************
 *			SPI_ARBITER.V																				*
 ********************************************************************************/
